#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/parport.h>
#include <linux/proc_fs.h>
#include <linux/types.h>
#include <linux/kernel.h>
#define MODULE_NAME "par2proc"

static struct proc_dir_entry *procdir;
static struct proc_dir_entry *file0;
static struct proc_dir_entry *file1;
static struct proc_dir_entry *file2;
static struct pardevice *ppdevice;
static struct parport *pp;
static int got_dev = 0;
static char reg[3];

static void ourparport_attach(struct parport*);
static void ourparport_detach(struct parport*);
static unsigned int parse_int(const char*, unsigned long);
static int proc_base0_read(char*, char**, off_t, int, int*, void*);
static int proc_base1_read(char*, char**, off_t, int, int*, void*);
static int proc_base2_read(char*, char**, off_t, int, int*, void*);
static int proc_base0_write(struct file*, const char*, unsigned long, void*);
static int proc_base1_write(struct file*, const char*, unsigned long, void*);
static int proc_base2_write(struct file*, const char*, unsigned long, void*);


static struct parport_driver ppdriver = {
	"par2proc",
	ourparport_attach,
	ourparport_detach,
	NULL
};

static void ourparport_attach(struct parport *port) {
	pp = port;
	ppdevice = parport_register_device(pp, "parport", NULL, NULL, NULL,
	                                   PARPORT_DEV_EXCL, NULL);
	if (!ppdevice) {
		printk("Could not open parport device.\n");
	} else {
		got_dev = 1;
		printk("Registered parport device.\n");
		if (parport_claim (ppdevice)) {
			printk("Could not get exlusive access to parport device.\n");
		} else {
			printk("Got exlusive access to parport device.\n");
			reg[0] = inb(pp->base);
			reg[1] = inb(pp->base+1);
			reg[2] = inb(pp->base+2);
			outb(0x0, pp->base);
			outb(0x0, pp->base+1);
			outb(0x0, pp->base+2);
		}
	}
}

static void ourparport_detach(struct parport *port) {
	if (got_dev == 1) {
		parport_unregister_device(ppdevice);
	}
	printk("Just lost parport device.\n");
}

static unsigned int parse_int(const char *buffer, unsigned long count) {
	int i = 0;
	unsigned char ret  = 0;
	unsigned char digits[4];
	if (!count) {
		printk("Too short\n");
		return -EINVAL;
	}
	if(count>4) {
		printk("Too long?\n");
		return -EINVAL;
	}
	if (copy_from_user(digits, buffer, count)) {
		printk("copy_from_user is not supposed to fail, but it just did.\n");
		return -EFAULT;
	}
	for (i = 0; i < (count-1); i++) {
		ret *=10;
		ret +=digits[i] - '0';
	}	
	return ret;
}

static int proc_base0_read(char* page, char** start, off_t off, int count,
                           int* eof, void* data) {

	int ret = 0;
	int len = 0;
	//MOD_INC_USE_COUNT;
	ret = inb(pp->base);
	len = sprintf(page, "%d\n", ret);
	//MOD_DEC_USE_COUNT;
	return len;
}

static int proc_base1_read(char* page, char** start, off_t off, int count,
                           int* eof, void* data) {

	int ret = 0;
	int len = 0;
	//MOD_INC_USE_COUNT;
	ret = inb(pp->base+1);
	len = sprintf(page, "%d\n", ret);
	//MOD_DEC_USE_COUNT;
	return len;
}

static int proc_base2_read(char* page, char** start, off_t off, int count,
                           int* eof, void* data) {

	int ret = 0;
	int len = 0;
	//MOD_INC_USE_COUNT;
	ret = inb(pp->base+2);
	len = sprintf(page, "%d\n", ret);
	//MOD_DEC_USE_COUNT;
	return len;
}

static int proc_base0_write(struct file* file, const char* buffer,
                            unsigned long count, void* data) {

	//MOD_INC_USE_COUNT;
	if (got_dev == 1) {
		outb(parse_int(buffer, count), pp->base);
	}
	//MOD_DEC_USE_COUNT;
	return count;
}

static int proc_base1_write(struct file* file, const char* buffer,
                            unsigned long count, void* data) {

	//MOD_INC_USE_COUNT;
	if (got_dev == 1) {
		outb(parse_int(buffer, count), pp->base+1);
	}
	//MOD_DEC_USE_COUNT;
	return count;
}

static int proc_base2_write(struct file* file, const char* buffer,
                            unsigned long count, void* data) {

	//MOD_INC_USE_COUNT;
	if (got_dev == 1) {
		outb(parse_int(buffer, count), pp->base+2);
	}
	//MOD_DEC_USE_COUNT;
	return count;
}

int init_module() {
	procdir = proc_mkdir("parport", 0);
	if (procdir == NULL) {
		printk("Can't create dir parport.\n");
		return -ENOMEM;
	}
	procdir->owner = THIS_MODULE;

	file0 = create_proc_entry("base+0", 0664, procdir);
    if (file0 == NULL) {
        printk("Can't create file parport/base+0.\n");
		remove_proc_entry("parport", 0);
        return -ENOMEM;
    }
	file0->owner = THIS_MODULE;
	file0->gid = 10; // wheel
	file0->read_proc = proc_base0_read;
	file0->write_proc = proc_base0_write;

	file1 = create_proc_entry("base+1", 0664, procdir);
    if (file1 == NULL) {
        printk("Can't create file parport/base+1.\n");
		remove_proc_entry("base+0", procdir);
		remove_proc_entry("parport", 0);
        return -ENOMEM;
    }
	file1->owner = THIS_MODULE;
	file1->gid = 10; // wheel
	file1->read_proc = proc_base1_read;
	file1->write_proc = proc_base1_write;

	file2 = create_proc_entry("base+2", 0664, procdir);
    if (file2 == NULL) {
        printk("Can't create file parport/base+0.\n");
		remove_proc_entry("base+1", procdir);
		remove_proc_entry("base+0", procdir);
		remove_proc_entry("parport", 0);
        return -ENOMEM;
    }
	file2->owner = THIS_MODULE;
	file2->gid = 10; // wheel
	file2->read_proc = proc_base2_read;
	file2->write_proc = proc_base2_write;

	parport_register_driver(&ppdriver);

	return 0;
}

void cleanup_module() {
	outb(reg[0], pp->base);
	outb(reg[1], pp->base+1);
	outb(reg[2], pp->base+2);
	parport_unregister_driver(&ppdriver);
	remove_proc_entry("base+2", procdir);
	remove_proc_entry("base+1", procdir);
	remove_proc_entry("base+0", procdir);
	remove_proc_entry("parport", 0);
}

MODULE_AUTHOR("Patrick 'Petschge' Kilian");
MODULE_DESCRIPTION("par2proc");
MODULE_LICENSE("GPL");
MODULE_VERSION("");
